from logging import info
from .finder import ChunkFinder
from .utils import AutoGet, filesizef2
from .descriptor import read_descriptor


class HtttpRangeSectorReader(AutoGet):
    def __init__(self, url: str, start: int, end: int, sector_size: int, rkw={}):
        self.start = start  # absolute start
        self.end = end  # absolute end
        self.url = url
        self.sector_size = sector_size
        self.rkw = rkw
        self.resp = None
        self.last: "list[int]" = [None, None]

    def at(self, pos_abs: int):
        from logging import info

        assert pos_abs >= self.start
        assert pos_abs < self.end
        B = self.sector_size
        q, r = divmod(pos_abs, B)
        if self.last[1] == q:
            info("Reuse %s %r", q, self.url)
            b = self.last[0]
            s = q * B
            return b, q, r, s, s + len(b)
        s = q * B
        o = s - self.start
        rkw = self.rkw
        rkw.setdefault("headers", {})["range"] = "bytes=%d-%d" % (o, o + B - 1)
        rkw["stream"] = True
        rkw["allow_redirects"] = True
        with self.http.get(self.url, **rkw) as h:
            c = h.status_code
            if c == 206:
                rh = getattr(h, "history", None)
                b = rh and rh[0].headers.get("Location", "").strip()
                if b:
                    info("Final URL %r --> %r", self.url, b)
                    self.url = b
                b = h.content
            elif c != 200:
                raise RuntimeError("Unexpected http status %r" % c)
            else:
                raise RuntimeError("Http byte range not supported %r" % c)
        self.last[0] = b
        self.last[1] = q
        return b, q, r, s, s + len(b)

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _get_http(self, name):
        from requests import session

        return session()


class RequestSectorReader(HtttpRangeSectorReader):
    def __init__(self, url, start, end, sector_size, rkw={}):
        super().__init__(url, start, end, sector_size, rkw)
        self.itr = None
        self.resp = None

    def at(self, pos_abs):
        from logging import info

        assert pos_abs >= self.start
        assert pos_abs < self.end
        B = self.sector_size
        q, r = divmod(pos_abs, B)
        if self.last[1] == q:
            info("Reuse %s %r", q, self.url)
            b = self.last[0]
            s = q * B
            return b, q, r, s, s + len(b)
        s = q * B
        o = s - self.start
        #
        if not self.itr:
            self.itr = self.get_iter(o)
        elif self.pos != o:
            assert self.pos % B == 0
            assert o > self.pos
            # if 0: #TODO: consume up to 'o'
            if (o - self.pos) < (16 * 1024 * 1024):
                info(
                    "Skip read %r->%r @%s %r",
                    (self.pos, self.start + self.pos, divmod(self.start + self.pos, B)),
                    (o, s, divmod(s, B)),
                    q,
                    self.url,
                )
                while self.pos < o:
                    next(self.itr)
                assert self.pos == o
            else:
                info(
                    "Reconnect %r->%r @%s %r",
                    (self.pos, self.start + self.pos, divmod(self.start + self.pos, B)),
                    (o, s, divmod(s, B)),
                    q,
                    self.url,
                )
                self.itr = self.get_iter(o)
            # assert(0)
        self.last[0] = b = next(self.itr)
        self.last[1] = q
        #
        return b, q, r, s, s + len(b)

    def iter_start(self):
        B = self.sector_size
        p = b""
        for c in self.resp.iter_content(B):
            p = p + c
            while len(p) >= B:
                b = p[:B]
                self.pos += len(b)
                yield b
                p = p[B:]
        if p:
            assert len(p) < B
            self.pos += len(p)
            yield p

    def get_iter(self, start):
        B = self.sector_size
        assert start % B == 0
        assert self.end > start
        from logging import info, warn

        rkw = self.rkw
        rkw.setdefault("headers", {})["range"] = "bytes=%d-%d" % (
            start,
            self.end - self.start - 1,
        )
        rkw["stream"] = True
        info(
            "Block fetch %r seek %r %r",
            self.url,
            (start, self.start + start, divmod(self.start + start, B)),
            rkw["headers"]["range"],
        )
        self.close()
        self.resp = r = self.http.get(self.url, **rkw)
        s = r.status_code
        i = self.iter_start()
        if s == 206:
            self.pos = start
        elif s != 200:
            raise RuntimeError("Unexpected http status %r", s)
        else:
            # raise RuntimeError("Http byte range not supported %r", s)
            warn("Byte range not supported %r walking up to %r", self.url, start)
            self.pos = 0
            while self.pos < start:
                next(i)
            assert self.pos == start
        return i

    def close(self):
        if self.resp:
            from logging import info

            info("Closing connection %r", self.url)
            self.resp = self.resp.close() and None


class FileSectorReader(object):
    def __init__(self, path, start, end, sector_size):
        self.start = start
        self.end = end
        self.path = path
        self.sector_size = sector_size
        self.last = [-1, -1]
        self.src = None

    def at(self, pos_abs):
        assert pos_abs >= self.start, f"{pos_abs} {self.start}..{self.end}"
        assert pos_abs < self.end, f"{pos_abs} {self.start}..{self.end}"
        B = self.sector_size
        q, r = divmod(pos_abs, B)
        if self.last[1] == q:
            from logging import debug

            debug("B reuse %s %r", q, self.path)
            b = self.last[0]
            s = q * B
            return b, q, r, s, s + len(b)
        s = q * B
        o = s - self.start
        if not self.src:
            from logging import debug

            debug("B open %r seek %s", self.path, o)
            self.src = open(self.path, "rb")
            self.src.seek(o)
        elif self.src.tell() != o:
            from logging import debug

            debug("B reseek %s->%s %r", self.src.tell(), o, self.path)
            self.src.seek(o)
            assert 0
        self.last[0] = b = self.src.read(B)
        self.last[1] = q
        return b, q, r, s, s + len(b)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if self.src:
            from logging import debug, info

            if getattr(self, "remove", None) is True:
                debug("B gone %r", self.path)
                self.src = self.src.close() and None
                from os import unlink

                try:
                    unlink(self.path)
                except Exception:
                    from sys import exc_info

                    info("Failed to delete %r, %s", self.path, str(exc_info()[1]))
            else:
                debug("B done %r", self.path)
                self.src = self.src.close() and None

    def remove_after(self):
        self.remove = True


def vfinal(f, because):
    if f.get("fh"):
        from logging import info

        info("Closing %r, %s", f["path"], because or "")
        f.pop("fh").close()
        assert "_oo_seek" not in f  # on open seek


def vopen(f):
    h = f.get("fh")
    if not h:
        from logging import info

        p = f["_path"]
        o = f.pop("_oo_seek")
        if o > 0:
            info("Appending %r @%s", p, filesizef2(o))
            m = "r+b"
        elif o < 0:
            from os.path import split, isdir

            d, n = split(p)
            if isdir(d):
                info("Creating %r", p)
            else:
                info("Creating %r at %r", n, d)
                from os import makedirs

                makedirs(d)
            m = "wb"
        else:
            info("Open %r", p)
            m = "r+b"
            # m = "wb"
        try:
            h = f["fh"] = open(p, m)
        except Exception:
            from logging import exception

            h = None
            exception("Failed ot open %r", p)
        else:
            (o > 0) and h.seek(o)
    assert "_oo_seek" not in f  # on open seek
    return h


class SerialConstructor(AutoGet):
    pwd: str

    def __init__(self, file):
        self.descriptor = read_descriptor(file)
        self.chunk_finders: "list[ChunkFinder]" = []

    def _get_sources(self):
        from itertools import chain

        return sorted(
            chain(*(c.items(self.descriptor["md5"], self.descriptor["size"]) for c in self.chunk_finders)),
            key=lambda x: x["offset"],
        )

    def _get_sector_size(self):
        from logging import info

        v = min(self.descriptor["chunks"].keys())
        info("Using sector_size %s", filesizef2(v))
        return v

    def _get_sector_size(self):
        from logging import info

        v = min(self.descriptor["chunks"].keys())
        info("Using sector_size %s", filesizef2(v))
        return v

    def _get_sector_map(self):

        return self.descriptor["chunks"][self.sector_size]

    def _get_pwd(self):

        from os.path import join
        from tempfile import gettempdir

        desc = self.descriptor
        pwd = join(gettempdir(), "%s-%s.tree" % (desc["md5"][:5], desc["size"]))
        info("Constructing at %t", pwd)
        return pwd

    def files_enum(self, start, end):
        from os import stat
        from os.path import join, exists
        from logging import info

        files = self.descriptor["files"]
        pos = 0

        def in_range(s, e):
            assert e >= s
            return s < end and e >= start and end > e
            # return (((s > end) and (end <= e)) or ((s >= start) and (start < e)))

        for f in files:
            # if "offset" in f:
            #     assert pos == f["offset"], f'NEQ {pos}, {f["offset"]} {f} '
            # else:
            #     f["offset"] = pos
            if "_path" not in f:
                f["_path"] = join(self.pwd, f["path"])
            size = f["size"]
            if size < 1:
                info("Zero size %r", f["path"])
                continue
            fh = f.get("fh")
            f_end = pos + size
            if fh:
                if in_range(pos, pos + fh.tell()):
                    yield f
                else:
                    vfinal(
                        f,
                        "@{} Outside {}-{}".format(filesizef2(fh.tell()), filesizef2(start), filesizef2(end)),
                    )
            elif exists(f["_path"]):
                current_size = stat(f["_path"]).st_size
                if current_size > size:
                    raise RuntimeError(
                        "Size %s expected not %s for %r"
                        % (
                            size,
                            current_size,
                            f["_path"],
                        )
                    )
                elif current_size == size:
                    if "_done" not in f:
                        f["_done"] = True
                        info("Done %r size %s", f["path"], filesizef2(current_size))
                elif in_range(pos, pos + current_size):  # current_size < size
                    f["_oo_seek"] = current_size
                    yield f
            elif in_range(pos, pos):
                f["_oo_seek"] = -1
                yield f
            pos += size

    def construct(self, **kwargs):
        from logging import error, info, exception
        from time import time
        from hashlib import md5 as mdnew

        B = self.sector_size
        S = self.sector_map
        url_sector_reader = kwargs.get("url_sector_reader") or RequestSectorReader

        def get_md(b):
            m = mdnew()
            m.update(b)
            return m.hexdigest()

        def block_name(v):
            _ = v.get("url")
            if _:
                return _
            return v["path"]

        def reader(v):
            _ = v.get("url", "")
            if _.startswith("file://"):
                from urllib.request import url2pathname
                from urllib.parse import urlparse

                v["path"] = url2pathname(urlparse(_)[2])
            elif _:
                return url_sector_reader(_, b_start, b_end, B)
            return FileSectorReader(v["path"], b_start, b_end, B)

        delete = kwargs.get("delete_finished_blocks")
        desc = self.descriptor
        check_hash = kwargs.get("check_hash")
        max_sectors = kwargs.get("max_sectors")
        total_size = desc["size"]
        for i, v in enumerate(self.sources):
            assert v["offset"] >= 0
            i_cur = None
            started = time()
            if "block_size" in v:
                assert v["offset"] % B == 0
                assert v["block_size"] > 0
                b_start = v["offset"]
                b_tail_size = total_size % v["block_size"]
                b_count = total_size // v["block_size"] + ((b_tail_size != 0) and 1 or 0)
                b_index = v["part_index"]
                assert b_count > 0
                assert b_index < b_count
                # b_end = b_start + v['block_size']
                b_end = b_start + (b_tail_size if b_index > 0 and (b_count - 1) == b_index else v["block_size"])
                b_fin = total_size if ((b_count - 1) == b_index) else (b_start + v["block_size"])
                assert b_count == 1 or b_fin == b_end, f"{b_fin} != {b_end}"
                info(
                    "B#%s %s-%s %s %s/%s %s",
                    i,
                    filesizef2(b_start),
                    filesizef2(b_end),
                    filesizef2(v["block_size"]),
                    b_index + 1,
                    b_count,
                    block_name(v),
                )
            else:
                b_start = v["offset"]
                b_end = v["end"]
                info("P#%s %s-%s", i, filesizef2(b_start), filesizef2(b_end))

            with reader(v) as sr:
                for j, f in enumerate(self.files_enum(b_start, b_end)):
                    h = vopen(f)
                    if not h:
                        continue
                    f_pos = f["offset"]
                    f_tell = h.tell()
                    f_start = f_pos + f_tell
                    f_end = f_pos + f["size"]
                    assert f_end > f_start
                    f_size = f_end - f_start
                    assert b_end > f_start
                    x_end = min(b_end, f_end)
                    info(
                        "F %s %s-%s-%s %r",
                        i,
                        filesizef2(f_pos),
                        filesizef2(f_start),
                        filesizef2(f_end),
                        f["path"],
                    )
                    i_cur = f_start
                    while i_cur < x_end:
                        try:
                            b, q, r, s, e = sr.at(i_cur)
                            md1 = S[q].get("md5")
                            if md1:
                                md0 = get_md(b)
                                if md0 != md1:
                                    error(
                                        "Wrong hash at sector %s expected %r got %r",
                                        q,
                                        md0,
                                        md1,
                                    )
                                    b = False
                        except Exception:
                            exception(
                                "B %r, F %r, C %r, %r",
                                (b_start, b_end),
                                (f_start, f_end),
                                (i_cur, x_end),
                                v.get("path"),
                            )
                            raise
                        if b is False:
                            break
                        b = b[r : min(e, x_end) - s]
                        # info("BUF %r", (r, s, e, x_end, i_cur, len(b), q))
                        if len(b) < 1:
                            raise RuntimeError("Empty sector %r" % ((r, s, e, x_end, i_cur, q),))
                        # print(" <W" , i_cur, len(b), f["fh"])
                        assert len(b)
                        h.write(b)
                        if max_sectors:
                            max_sectors -= 1
                            if max_sectors <= 0:
                                break
                        i_cur += len(b)
                        now = time()
                        if (now - started) > 30:
                            started = now
                            info(
                                "F %s %s-%s-%s",
                                i,
                                filesizef2(f_start),
                                filesizef2(i_cur),
                                filesizef2(f_end),
                            )
                    if i_cur == f_end:
                        vfinal(f, "At end")
                        if check_hash:
                            vcheck(f)
                if i_cur == b_end:
                    if delete:
                        sr.remove = True
                    # info("Delete? {!r}".format(v['path']))
        #
        for f in self.files_enum(self.descriptor["size"], self.descriptor["size"] + 1):
            pass  # close the last file
        #

    def check(self, chunk_size=25 * 1024 * 1024):
        from os.path import join, exists
        from os import stat
        from hashlib import md5 as mdnew

        files = self.descriptor["files"]
        n = max(len(str(f["size"])) for f in files)
        i = len(str(len(files)))
        mask = "{0:" + str(i) + "d} {md5} {size:" + str(n) + "d} {path}"
        # print(mask)
        c_not_found = 0
        c_size = 0
        c_hash = 0

        def range_clue(f, size_partial=0):
            n = f["offset"] + f["size"]
            last_index = (n // chunk_size) + (1 if (n % chunk_size) else 0) - 1
            n = f["offset"] + size_partial
            first_index = n // chunk_size
            return "-".join(str(x) for x in sorted(set([first_index, last_index])))

        for i, f in enumerate(files):
            full = join(self.pwd, f["path"])
            print(mask.format(i, **f))
            if not exists(full):
                # print(f"\t!NotFound offset:{f['offset']}")
                print(f"{f['offset']:>16} !! NotFound {range_clue(f)}")
                c_not_found += 1
                continue
            st = stat(full)
            if st.st_size != f["size"]:
                c_size += 1
                print(f"{f['offset']:>16} !! Size:{st.st_size} diff:{f['size']-st.st_size} {range_clue(f, st.st_size)}")
                continue
            md = mdnew()
            md.update(open(full, "rb").read())
            if md.hexdigest() != f["md5"]:
                print(f"{f['offset']:>16} !! MD {md.hexdigest()}")
                c_hash += 1
                continue
            print(f"{f['offset']:>16} OK")
        if (c_not_found + c_size + c_hash) == 0:
            print("No Errors")

    def available_ranges(self):
        from os import stat
        from os.path import exists
        from logging import info

        files = self.descriptor["files"]
        pos = 0
        for i, f in enumerate(files):
            if "offset" in f:
                assert pos == f["offset"]
            else:
                f["offset"] = pos
            size = f["size"]
            if size < 1:
                info("Zero size %r", f["path"])
                continue
            p = self.get_full_path(f)
            if exists(p):
                current_size = stat(p).st_size
                if current_size > size:
                    raise RuntimeError(
                        "Size %s expected not %s for %r"
                        % (
                            size,
                            current_size,
                            p,
                        )
                    )
                else:
                    yield (f["offset"], f["offset"] + current_size - 1)
            pos += size

    def non_available_ranges(self):
        from os import stat
        from os.path import exists
        from logging import info

        files = self.descriptor["files"]
        pos = 0
        for i, f in enumerate(files):
            if "offset" in f:
                assert pos == f["offset"]
            else:
                f["offset"] = pos
            size = f["size"]
            if size < 1:
                info("Zero size %r", f["path"])
                continue
            p = self.get_full_path(f)
            first_byte = f["offset"]
            last_byte = f["offset"] + size - 1
            if not exists(p):
                print(f"not exists({p!r})")
                yield (first_byte, last_byte)
            else:
                current_size = stat(p).st_size
                if current_size > size:
                    raise RuntimeError(
                        "Size %s expected not %s for %r"
                        % (
                            size,
                            current_size,
                            p,
                        )
                    )
                elif current_size < size:
                    print(f"current_size:{current_size} size:{size} {p}")
                    yield (first_byte + current_size, last_byte)
            pos += size

    def get_full_path(self, item):
        p = item.get("_path")
        if not p:
            from os.path import join

            p = item["_path"] = join(self.pwd, item["path"])
        return p


def vcheck(f, chunksize=131072):
    from hashlib import md5 as mdnew
    from os import stat
    from logging import info

    p = f["_path"]
    st = stat(p)
    if st.st_size == f["size"]:
        info("Checking hash of %r", p)
        md = mdnew()
        with open(p, "rb") as h:
            b = h.read(chunksize)
            while b:
                md.update(b)
                b = h.read(chunksize)
        if md.hexdigest() == f["md5"]:
            print(
                "Ok {!r} {!r}".format(
                    p,
                    [f["md5"], f["size"]],
                )
            )
        else:
            print("Invalid Hash {!r} {!r}".format(p, [f["md5"], md.hexdigest()]))
    else:
        print("Invalid Size {!r} {!r}".format(p, [f["size"], st.st_size, st.st_size - f["size"]]))


class Sink(object):
    def __init__(self, file):
        self.descriptor = read_descriptor(file)
        self.chunk_finders = []

    def construct(self, out, **kwargs):
        from itertools import chain
        from logging import info
        from hashlib import md5 as mdnew

        l = sorted(
            chain(*(c.items(self.descriptor["md5"], self.descriptor["size"]) for c in self.chunk_finders)),
            key=lambda x: x["offset"],
        )

        desc = self.descriptor
        B = min(self.descriptor["chunks"].keys())
        S = self.descriptor["chunks"][B]

        info("md5 %r", desc["md5"])
        info("size %r", desc["size"])
        url_sector_reader = kwargs.get("url_sector_reader") or RequestSectorReader
        total_size = desc["size"]

        def get_md(b):
            m = mdnew()
            m.update(b)
            return m.hexdigest()

        def block_name(v):
            _ = v.get("url")
            if _:
                return _
            return v["path"]

        def reader(v):
            _ = v.get("url", "")
            if _.startswith("file://"):
                from urllib.request import url2pathname
                from urllib.parse import urlparse

                v["path"] = url2pathname(urlparse(_)[2])
            elif _:
                return url_sector_reader(_, b_start, b_end, B)
            return FileSectorReader(v["path"], b_start, b_end, B)

        def enum():
            pos = 0
            for i, v in enumerate(l):
                if not v:
                    continue
                b_start = v["offset"]
                b_tail_size = total_size % v["block_size"]
                b_count = total_size // v["block_size"] + ((b_tail_size != 0) and 1 or 0)
                b_index = v["part_index"]
                assert b_count > 0
                assert b_index < b_count
                b_end = total_size if ((b_count - 1) == b_index) else (b_start + v["block_size"])
                if pos < b_start:
                    continue
                elif pos >= b_end:
                    s[i] = 0
                    continue
                elif pos == b_start:
                    pos = b_end
                    yield b_start, b_end, v
                else:
                    # pos >= off
                    # pos < end
                    pass
                    yield b_start, b_end, v

        def get_reader(v, start, end, B):
            _ = v.get("url")
            if _:
                return url_sector_reader(_, start, end, B)
            return FileSectorReader(v["path"], start, end, B)

        e = -1
        for i, (s, e, v) in enumerate(enum()):
            info("%r\t%r\t%r\t%r", i, s, e, block_name(v))
        if e == desc["size"]:
            for i, (b_start, b_end, v) in enumerate(enum()):
                info("%r\t%r\t%r\t%r", i, s, e, block_name(v))
                with get_reader(v, b_start, b_end, B) as sr:
                    pos = b_start
                    while pos < b_end:
                        j = "."
                        b, q, r, s, e = sr.at(pos)
                        md1 = S[q].get("md5")
                        if md1:
                            md0 = get_md(b)
                            j = "o" if md0 == md1 else "x"
                        if out:
                            out.write(b)
                        else:
                            print(j, end="", sep="")
                        pos += B



import unittest


class Test(unittest.TestCase):
    def test_size_and_offsets(self):
        s = total_size = 11
        b = block_size = 3

        def total_blocks():
            return s // b + (1 if (s % b) else 0)

        def last_index():
            return s // b + (0 if (s % b) else -1)

        def last_index_off(s, B, o):
            return ((o + s) // B) + (0 if ((o + s) % B) else -1)

        def past_index_off(s, B, o):
            return ((o + s) // B) + (((o + s) % B) and 1 or 0)

        def first_index_off(o, B):
            return o // B

        # def first_index_off(o, B):
        #     return ((o + s) // B) - (o // B)

        for x, i in [(11, 4), (9, 3), (1, 1), (2, 1), (3, 1), (4, 2)]:
            s = total_size = x
            self.assertEqual(
                total_blocks(),
                i,
                f"total blocks total_size={total_size}, block_size={block_size}",
            )

        for x, i in [(4, 1), (2, 0), (5, 1), (6, 1)]:
            s = total_size = x
            self.assertEqual(
                last_index(),
                i,
                f"last index total_size={total_size}, block_size={block_size}",
            )

        s = total_size = 11
        b = block_size = 3
        # print(first_index_off(4, 3, 2))

        for offset, size, index_first, index_last in [
            (0, 2, 0, 0),
            (0, 1, 0, 0),
            (0, 3, 0, 0),
            (0, 4, 0, 1),
            (0, 5, 0, 1),
            (0, 6, 0, 1),
            (2, 1, 0, 0),
            (2, 4, 0, 1),
            (4, 4, 1, 2),
            (3, 6, 1, 2),
            (6, 3, 2, 2),
            (0, 0, 0, -1),
        ]:
            self.assertEqual(
                first_index_off(offset, block_size),
                index_first,
                f"first_index_off {(offset, size, index_first, index_last)!r}, {(total_size, block_size)!r}",
            )
            x = last_index_off(size, block_size, offset)
            self.assertEqual(
                x,
                index_last,
                f"last_index_off {(offset, size, index_first, index_last)!r}, {(total_size, block_size)!r}",
            )
            self.assertEqual(
                past_index_off(size, block_size, offset),
                x + 1,
                f"past_index_off {(offset, size, index_first, index_last)!r}, {(total_size, block_size)!r}",
            )

    def test_list_ranges(self):
        from blob_descriptor.utils import list_ranges

        for indexes, ranges in [
            [[0, 1, 2, 3, 4], [[0, 4]]],
            [[0, 2, 4], [[0, 0], [2, 2], [4, 4]]],
            [[0, 2, 3, 4], [[0, 0], [2, 4]]],
            [[0, 1, 2, 4], [[0, 2], [4, 4]]],
            [[0, 1, 4, 5], [[0, 1], [4, 5]]],
        ]:
            self.assertEqual(
                ranges,
                list_ranges(indexes),
                f"list_ranges {indexes!r}, {ranges!r}",
            )


if __name__ == "__main__":
    unittest.main()



import hashlib
import unittest
import os
import tempfile
from hashlib import md5
from blob_descriptor import RegSource


class TestRegSource(unittest.TestCase):
    def setUp(self):
        # Create a temporary file with known content
        self.test_content = b"hello world" * 1000  # 11,000 bytes
        self.tmp = tempfile.NamedTemporaryFile(delete=False)
        self.tmp.write(self.test_content)
        self.tmp.close()
        self.regsource = RegSource(self.tmp.name)

    def tearDown(self):
        os.unlink(self.tmp.name)  # Cleanup

    def test_size(self):
        self.assertEqual(self.regsource.size, len(self.test_content))

    def test_md5(self):
        expected = md5(self.test_content).hexdigest()
        self.assertEqual(self.regsource.md5, expected)

    def test_iter_chunks(self):
        chunks = list(self.regsource.iter_chunks(1024))
        joined = b"".join(chunks)
        self.assertEqual(joined, self.test_content)
        self.assertTrue(all(len(chunk) <= 1024 for chunk in chunks))

    def test_unknown_attribute(self):
        with self.assertRaises(AttributeError):
            _ = self.regsource.unknown_attribute


class TestRegSource2(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Create a temporary file for testing
        cls.temp_file = tempfile.NamedTemporaryFile(delete=False)
        cls.test_data = b"This is some test data for RegSource"
        cls.temp_file.write(cls.test_data)
        cls.temp_file.close()

        # Calculate expected MD5
        cls.expected_md5 = hashlib.md5(cls.test_data).hexdigest()
        # Get expected size
        cls.expected_size = len(cls.test_data)

    @classmethod
    def tearDownClass(cls):
        # Clean up the temporary file
        os.unlink(cls.temp_file.name)

    def test_initialization(self):
        # Test initialization with file path
        source = RegSource(file=self.temp_file.name, path="test_path")
        self.assertEqual(source.file, self.temp_file.name)
        self.assertEqual(source.path, "test_path")

        # Test initialization without path
        source = RegSource(file=self.temp_file.name)
        # self.assertEqual(source.path, os.path.basename(self.temp_file.name))

    def test_size_property(self):
        source = RegSource(file=self.temp_file.name)
        # Test size property (lazy loading)
        self.assertEqual(source.size, self.expected_size)
        # Test that it's cached
        self.assertTrue("size" in source.__dict__)

    def test_md5_property(self):
        source = RegSource(file=self.temp_file.name)
        # Test md5 property (lazy loading)
        self.assertEqual(source.md5, self.expected_md5)
        # Test that it's cached
        self.assertTrue("md5" in source.__dict__)

    def test_iter_chunks(self):
        source = RegSource(file=self.temp_file.name)
        block_size = 10  # Small block size for testing

        # Test iter_chunks yields correct data
        chunks = list(source.iter_chunks(block_size))
        combined = b"".join(chunks)
        self.assertEqual(combined, self.test_data)

        # Test chunk sizes (last chunk may be smaller)
        for chunk in chunks[:-1]:
            self.assertEqual(len(chunk), block_size)
        self.assertLessEqual(len(chunks[-1]), block_size)

    def test_calc_md5(self):
        source = RegSource(file=self.temp_file.name)
        # Test calc_md5 method directly
        self.assertEqual(source.calc_md5(), self.expected_md5)

        # Test with custom block size
        self.assertEqual(source.calc_md5(block_size=5), self.expected_md5)

    def test_repr(self):
        source = RegSource(file=self.temp_file.name, path="test_path")
        # Test string representation
        self.assertEqual(repr(source), f"RegSource('test_path')")

    def test_nonexistent_file(self):
        # Test behavior with non-existent file
        with self.assertRaises(FileNotFoundError):
            source = RegSource(file="nonexistent_file")
            _ = source.size  # Should raise FileNotFoundError when trying to get size

    def test_invalid_attribute(self):
        source = RegSource(file=self.temp_file.name)
        # Test accessing invalid attribute
        with self.assertRaises(AttributeError):
            _ = source.invalid_attr


if __name__ == "__main__":
    unittest.main()



from configparser import ConfigParser
from .finder import ChunkFinder
from .main import Main, arg, flag
from .utils import filesizef2, input_file, list_ranges


class Verify(Main):
    descriptor: str = arg("Descriptor file")
    chunks: "list[str]" = arg("Chunks for the descriptor", nargs="*")
    output_dir: str = flag("d", "Output files to DIRectory", default=".", metavar="DIR")
    check_hash: bool = flag("check-hash", "Verify hashes after assemble")
    search_dirs: "list[str]" = flag("search", "s", "Search DIR for chunks", metavar="DIR", default=[])

    config: "ConfigParser"
    config_path: str
    chunk_paths: "dict[str, str]"

    def _get_config_path(self):
        from os import environ
        from os.path import expanduser

        return environ.get("BLOB_DESCRIPTOR_CONF") or expanduser("~/.config/blob_descriptor.conf")

    def _get_config(self):
        from configparser import ConfigParser
        from os.path import exists

        config = ConfigParser()
        if exists(self.config_path):
            config.read(self.config_path)
        return config

    def _get_chunk_paths(self):
        from os.path import expanduser, isdir, abspath

        def enum():
            try:
                dirs = self.config.get("blob_descriptor", "search").splitlines()
            except Exception:
                pass
            else:
                for v in dirs:
                    if v:
                        v = expanduser(v)
                        if isdir(v):
                            yield v
            try:
                dirs = self.search_dirs or ()
            except Exception:
                pass
            else:
                for v in dirs:
                    if v and isdir(v):
                        yield v

        return dict((abspath(x), x) for x in enum())  # dict as ordered set

    def find_chunks(self):
        from .finder import ChunkFinder
        from os.path import isdir

        finder = ChunkFinder()
        for a in self.chunks:
            if isdir(a):
                self.chunk_paths[a] = a
            elif "://" in a:
                # print('check_url', a)
                finder.check_url(a)
            else:
                finder.check_file(a)
        for d in self.chunk_paths:
            # print("SD", d)
            finder.search(d)
        return finder

    def ready(self) -> None:
        from logging import basicConfig
        from os import environ

        format = environ.get("LOG_FORMAT", "%(levelname)s: %(message)s")
        level = environ.get("LOG_LEVEL", "INFO")
        basicConfig(format=format, level=level)
        return super().ready()

    def descriptor_files(self):
        from os.path import exists, isabs, join, isfile

        path = self.descriptor
        if not path:
            return
        elif path == ":":
            path = input_file("Enter bd file")
        elif not exists(path) and not isabs(path):
            for d in self.chunk_paths.keys():
                a = join(d, path)
                if isfile(a):
                    yield a
        if exists(path):
            yield path
        else:
            raise RuntimeError(f"Not found {path}")

    def VERIFY(self, path: str, finder: "ChunkFinder"):
        from os.path import exists
        from os import stat
        from hashlib import md5 as mdnew
        from logging import info
        from .construct import SerialConstructor

        const = SerialConstructor(path)
        const.pwd = self.output_dir
        const.chunk_finders.append(finder)

        desc = const.descriptor
        total_size = desc["size"]
        block_sizes: set[int] = set(x for x in desc["chunks"].keys())
        files = desc["files"]

        if 1:
            block_sizes_map: "dict[int, dict[str, object]]" = dict((s, {}) for s in block_sizes)
            # print(block_sizes_map)
            for s in block_sizes:
                total_blocks = total_size // s + (1 if (total_size % s) else 0)
                block_sizes_map[s]["index_map"] = index_map = dict((i, None) for i in range(total_blocks))

        def tick_byte_range(a, b, f, val=True):
            for s, m in block_sizes_map.items():
                # print("\t", sorted(set([first_index, last_index])))
                last_index = ((b + 1) // s) + (0 if ((b + 1) % s) else -1)
                first_index = a // s  # + (1 if (a % s) else 0)
                assert first_index <= last_index
                i = first_index
                while i <= last_index:
                    m["index_map"][i] = val
                    i += 1

        print(f"{desc['md5']} {path}")
        print(
            f"     {len(desc['files'])} files {desc['size']} bytes",
            ",".join(f"{filesizef2(s)}x{len(m)}" for s, m in desc["chunks"].items()),
            "chunks",
        )
        N = len(str(total_size))

        pos = 0
        for i, f in enumerate(files):
            if "offset" in f:
                assert pos == f["offset"]
            else:
                f["offset"] = pos
            size = f["size"]
            if size < 1:

                info("Zero size %r", f["path"])
                continue
            p = const.get_full_path(f)
            first_byte = f["offset"]
            past_byte = f["offset"] + size
            last_byte = f["offset"] + size - 1
            end_byte = -1

            print(f"  {f['md5']} {f['path']}")

            def enum():
                yield f"%{N}d" % (f["offset"],)
                yield f"%{N}db" % size
                for s, m in desc["chunks"].items():
                    last_index = (past_byte // s) + (0 if (past_byte % s) else -1)
                    first_index = f["offset"] // s
                    total_blocks = last_index - first_index + 1
                    if first_index == last_index:
                        yield (f"{filesizef2(s)}({first_index})")
                    elif end_byte > 0:
                        end_index = end_byte // s
                        yield (f"{filesizef2(s)}({first_index}-{end_index}-{last_index})")
                    else:
                        yield (f"{filesizef2(s)}({first_index}-{last_index})")
                    assert first_index >= 0
                    assert last_index >= first_index
                    assert total_blocks > 0
                    assert (
                        (last_index - first_index) + 1
                    ) == total_blocks, f"{first_index}-{last_index}, {total_blocks}, {f!r}"

            indc = ""

            if exists(p):
                current_size = stat(p).st_size
                if current_size > size:
                    raise RuntimeError(
                        "Size %s expected not %s for %r"
                        % (
                            size,
                            current_size,
                            p,
                        )
                    )
                elif current_size < size:
                    end_byte = first_byte + current_size
                    tick_byte_range(end_byte, last_byte, f)
                    indc = f"PARTIAL({filesizef2(current_size)})"
                else:
                    # TODO: check hash
                    indc = "COMPLETE"
                    if self.check_hash:
                        md = mdnew()
                        with open(p, "rb") as fh:
                            md.update(fh.read())
                        if md.hexdigest() == f["md5"]:
                            indc = f"VALID"
                        else:
                            indc = f"INVALID Hash {md.hexdigest()!r}"
            else:
                tick_byte_range(first_byte, last_byte, f)
                indc = "ABSENT"

            print(f"    ", *enum(), indc)

            pos += size
        # From the found checksum (Not working)
        for s, m in block_sizes_map.items():
            parts = [
                x
                for f in const.chunk_finders
                for x in f.all()
                if desc["md5"].startswith(x["checksum"]) and x["block_size"] == s
            ]
            for x in parts:
                i = x.get("part_index", -1)
                if i > -1:
                    # TODO: check hash
                    print(i, x["path"], m["index_map"][i])
                    m["index_map"][i] = None
        # setup missing
        for s, m in block_sizes_map.items():
            if any(x for i, x in m["index_map"].items() if x):
                m["missing"] = list_ranges(sorted(i for i, x in m["index_map"].items() if x))

        head = None
        for s, m in block_sizes_map.items():
            d = m.get("missing")
            if not d:
                continue
            elif not head:
                print("Missing:")
                head = True
            # exclude chunks
            n = sum(((b - a) + 1) for a, b in d)

            print(
                " ",
                f"{filesizef2(s)} x{n}",
                ",".join(f"{a}{'' if a==b else '-%s'%(b)}" for (a, b) in m["missing"]),
                filesizef2(n * s),
            )

    def start(self) -> None:
        finder = self.find_chunks()
        for f in self.descriptor_files():
            self.VERIFY(f, finder)



import unittest


def make_chunks():
    import os, hashlib

    h_all = hashlib.new("md5")

    m = {}
    a = []
    blobs = [os.urandom(512 * 1024) for i in range(5)]
    blobs.append(os.urandom(1024 // 4))

    for i, blob in enumerate(blobs):
        h_cur = hashlib.new("md5")
        h_cur.update(blob)

        a.append(dict(index=i, blob=blob, hash=h_cur.hexdigest()))
        h_all.update(blob)

    m["hash"] = h_all.hexdigest()
    m["blob"] = b"".join(x.pop("blob") for x in a)
    m["512"] = a
    return m


class Test(unittest.TestCase):
    def test_cmd(self):
        import pprint
        from tempfile import mkdtemp, TemporaryDirectory, NamedTemporaryFile
        from os import chdir, urandom
        from subprocess import run

        with TemporaryDirectory() as temp_dir:
            self._create(temp_dir)
        with TemporaryDirectory() as temp_dir:
            self._create(temp_dir, "files")

    def _create(self, wdir, var=None):
        import pprint, random, os
        from tempfile import mkdtemp, TemporaryDirectory, NamedTemporaryFile
        from os import chdir, urandom, listdir
        from subprocess import run, check_output
        from random import uniform

        chdir(wdir)

        def shell(s, **kwargs):
            return run(s, shell=True, check=True, **kwargs)

        cdata = make_chunks()

        if var == "files":
            apple = urandom(int(uniform(1024 * 1024, 1024 * 1024 * 3)))
            with open("apple", "wb") as w:
                w.write(apple)
            banana = urandom(int(uniform(1024 * 1024, 1024 * 1024 * 3)))
            with open("banana", "wb") as w:
                w.write(banana)
            carrot = urandom(int(uniform(1024 * 1024, 1024 * 1024 * 3)))
            with open("carrot", "wb") as w:
                w.write(carrot)
            shell(
                "python3 -m blob_descriptor create -o file.bd --cw 1m,1M carrot apple banana",
            )
            total_size = sum(len(x) for x in [apple, banana, carrot])

        else:
            tf = NamedTemporaryFile()
            tf.write(cdata["blob"])
            tf.seek(0)
            total_size = len(cdata["blob"])

            try:
                shell(
                    "python3 -m blob_descriptor create -o file.bd --cs 700k --cw 1m,1M stdin:///input",
                    stdin=tf,
                )
            except Exception:
                cdata.pop("blob")

                pprint.pprint(cdata)
                raise
        shell(
            "ls -1shR .",
        )

        self.assertEqual(sum(os.stat(f"1M/{x}").st_size for x in os.listdir("1M")), total_size)

        shell(
            "python3 -m blob_descriptor verify file.bd",
        )


if __name__ == "__main__":
    unittest.main()



from logging import info
from typing import IO
from .descriptor import write_descriptor


class AutoGet:

    def __getattr__(self, name: str) -> object:
        if not name.startswith("_get_"):
            f = getattr(self, f"_get_{name}", None)
            if f:
                setattr(self, name, None)
                v = f()
                setattr(self, name, v)
                return v
        try:
            m = super().__getattr__
        except AttributeError:
            raise AttributeError(f"{self.__class__.__name__} has no attribute {name}") from None
        else:
            return m(name)


class ChunkLister(object):
    """Handles splitting binary data into chunks and tracking chunk metadata."""

    def __init__(self, chunk_size: int):
        from hashlib import md5
        from base64 import b64encode

        self.encoder = lambda b: b64encode(b).decode()
        self.mdnew = md5
        self.observers = []  #  List of observer objects to notify
        self.chunks: "list[dict[str, str | int]]" = []  # List of chunk metadata dictionaries
        self.current_md = self.mdnew()  # Current hash object being built
        self.current_n = 0  # Bytes accumulated in current chunk
        self.current_index = 0  # Index of current chunk
        self.current_prefix = None  # First few bytes of current chunk
        self.chunk_size = chunk_size  # Size of chunks in bytes

    def update(self, data: bytes):
        """Process incoming data and distribute to chunks"""
        B = self.chunk_size
        pos = 0
        part = data[pos : B - self.current_n]
        while part:
            if self.current_n == 0:
                assert self.current_prefix is None
                self.current_prefix = part[0:6]
                self.notifyObservers("in")
            l = len(part)
            self.current_md.update(part)
            self.notifyObservers("data", part)
            self.current_n += l
            pos += l
            if self.current_n == B:
                self.push()
            else:
                assert self.current_n < B
                assert self.current_n > 0
            part = data[pos : B - self.current_n]

    def push(self):
        """Finalize current chunk and add to chunks list."""
        d = dict(
            md5=self.current_md.hexdigest(),
            prefix=self.encoder(self.current_prefix),
            size=self.current_n,
            index=self.current_index,
        )
        self.notifyObservers("out", d)
        self.chunks.append(d)
        self.current_index += 1
        self.current_md = self.mdnew()
        self.current_n = 0
        self.current_prefix = None

    def get_chunks(self):
        """Get all completed chunks, finalizing any in-progress chunk.

        Returns:
            list: List of chunk metadata dictionaries
        """
        if self.current_prefix is not None:
            self.push()
        return self.chunks

    def notifyObservers(self, *args):
        """Notify all observers of chunk events.

        Args:
            *args: Event arguments (event_type, data)
        """
        for o in self.observers:
            o.update(self, *args)


class BlobDescriptor(object):
    def __init__(self):
        self.files: list[Source] = []  # List of Source objects to process
        self.no_duplicates = None  # hether to check for duplicates
        self.observers = []  # List of observer objects
        self.chunk_writers: "list[int | ChunkWriter]" = [
            512 * 1024,
        ]  # List of chunk writer configurations

    def make_descriptor(self, **kwargs):
        """Generate descriptor dictionary for all files."""
        block_size = kwargs.get("block_size", 16384)
        desc = kwargs.setdefault("descriptor", {})
        files = desc.setdefault("files", [])
        cwmap: "dict[int, ChunkLister]" = {}
        for cw in self.chunk_writers:
            if isinstance(cw, ChunkWriter):
                cs = cw.chunk_size
            else:
                cs = cw
                cw = None
            if cs in cwmap:
                cl = cwmap[cs]
            else:
                cl = cwmap[cs] = ChunkLister(cs)
            if cw is not None:
                if cw not in cl.observers:
                    cl.observers.append(cw)
                if cw not in self.observers:
                    self.observers.append(cw)

        chunk_gen = cwmap.values()
        # self.notifyObservers('chunk_listers', chunk_gen)
        from hashlib import md5 as mdnew

        total_hash = mdnew()
        total_size = 0
        for f in self.iter_files():
            _file_size = self.no_duplicates and getattr(f, "size", None)
            # info("DUP check %r", [self.no_duplicates, _file_size])
            if _file_size and any(1 for v in files if v["size"] == _file_size):
                # info("DUP _file_size %r", _file_size)
                _md5 = getattr(f, "md5", None)
                if _md5:
                    # info("DUP _md5 %r", _md5)
                    _item = next(
                        filter(
                            (lambda v: v["size"] == _file_size and v["md5"] == _md5),
                            files,
                        ),
                        0,
                    )
                    if _item:
                        info("DUP _item %r", _item)
                        files.append(dict(_item, path=f.path))
                        continue

            file_hash = mdnew()
            file_size = 0
            offset = total_size
            for b in f.iter_chunks(block_size):
                l = len(b)
                file_size += l
                total_size += l
                file_hash.update(b)
                total_hash.update(b)
                for g in chunk_gen:
                    g.update(b)
            md5 = file_hash.hexdigest()
            if not self.no_duplicates:
                _file_size = getattr(f, "size", None)
                if _file_size is not None and file_size != _file_size:
                    raise RuntimeError("Unexpected size %r" % ((_file_size, file_size),))
                else:
                    _md5 = getattr(f, "md5", None)
                    if _md5 is not None and md5 != _md5:
                        raise RuntimeError(f"Unexpected hash {(_md5, md5)!r} for {f!r}")
            files.append(dict(md5=md5, size=file_size, path=f.path, offset=offset))
        desc["md5"] = total_hash.hexdigest()
        desc["size"] = total_size
        desc["chunks"] = dict((g.chunk_size, g.get_chunks()) for g in chunk_gen)
        return desc

    def format_descriptor_path(self, dir="", prefix="", suffix="", stem="desc", ext=".bd"):
        """Generate path for descriptor file"""

        if not dir:

            from tempfile import gettempdir

            dir = gettempdir()

        from os.path import join

        return join(dir, prefix + stem + suffix + ext)

    def save(self, path: "str | dict", **kwargs):
        """Save descriptor to file"""
        desc = self.make_descriptor(**kwargs)
        if not path:
            from os.path import join
            from tempfile import gettempdir

            path = join(gettempdir(), "%s_%s_desc" % (desc["md5"][:5], desc["size"]))
        elif isinstance(path, dict):
            if path.get(""):
                path = path.get("")
            else:
                f = path.pop("stemf", None)
                hash = desc["md5"][:5]
                size = desc["size"]
                if f:
                    path = self.format_descriptor_path(stem=f.format(hash=hash, size=size), **path)
                else:
                    path = self.format_descriptor_path(stem="{}_{}".format(hash, size), **path)
        assert isinstance(path, str)

        info("Saving descriptor {!r}".format(path))
        write_descriptor(desc, path, chunk_size=False, chunk_index=False) and self.notifyObservers(
            "descriptor_saved", desc, path
        )

    def add_file(self, file, path="", **kwargs):
        """Add file to be processed.
        Args:
            file (str|file-like): File path or object
            path (str): Alternate path to store in descriptor
        """
        if not path:
            from os.path import basename

            path = basename(file)
        if self.no_duplicates:
            self.files.append(RegSource(file=file, path=path))
        else:
            self.files.append(FileSource(file=file, path=path))

    def iter_files(self):
        """Iterate through all registered files."""
        for f in self.files:
            yield f

    def notifyObservers(self, *args):
        """Notify all observers of descriptor events."""
        for o in self.observers:
            o.update(self, *args)

    def add_tree(self, path, **kwargs):
        """Recursively add all files in directory tree.

        Args:
            path (str): Root directory path
            **kwargs: Filter options:
                excludes: Patterns to exclude
                includes: Patterns to include
        """
        from pathlib import Path
        from logging import info

        top = Path(path)
        excludes = kwargs.get("excludes")
        includes = kwargs.get("includes")
        if self.no_duplicates:
            Source = RegSource
        else:
            Source = FileSource

        for sub in top.rglob("*"):
            if sub.is_dir():
                continue
            if excludes:
                if any(m for m in excludes if sub.match(m)):
                    continue
            if sub and includes:
                if any(m for m in includes if not sub.match(m)):
                    continue
            if sub:
                r = sub.relative_to(top)
                info("Add %r as %r", str(sub), str(r))
                self.files.append(Source(file=str(sub), path=str(r)))


from hashlib import md5
from os import environ, stat


class Source(AutoGet):
    """Base class for data sources."""

    path: str  # Logical path for the source
    md5: "str | None"
    size: "int | None"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.path!r})"

    def iter_chunks(self, block_size: int):
        yield b""


class RegSource(Source):
    """Source for regular files with on-demand hash calculation."""

    def __init__(self, file="", path=""):
        self.file = file
        self.path = path

    def _get_size(self):
        return stat(self.file).st_size

    def _get_md5(self):
        return self.calc_md5()

    def iter_chunks(self, block_size: int):
        with open(self.file, "rb") as h:
            b = h.read(block_size)
            while b:
                yield b
                b = h.read(block_size)

    def calc_md5(self, block_size=131072):
        m = md5()
        with open(self.file, "rb") as h:
            b = h.read(block_size)
            while b:
                m.update(b)
                b = h.read(block_size)
        return m.hexdigest()


class FileSource(Source):
    """Source for file-like objects with pre-known metadata."""

    def __init__(self, file, path: str = None, size: "int | None" = None, md5: str = ""):
        self.file = file
        self.path = path
        if size is not None:
            self.size = size
        if md5:
            self.md5 = md5

    def iter_chunks(self, block_size: int):
        if hasattr(self.file, "read"):
            h = self.file
        else:
            h = open(self.file, "rb")
        if h:
            b = h.read(block_size)
            while b:
                yield b
                b = h.read(block_size)


class URLSource(Source):
    """Source for remote URLs."""

    def __init__(self, url, path="", size=-1, md5=""):
        self.url = url
        from urllib.parse import urlparse

        u = urlparse(url.strip("/"))
        self.path = path or u.path.split("/")[-1]

    def iter_chunks(self, block_size: int):
        from requests import get as fetch

        return fetch(self.url).iter_content(block_size)


def mask1(md5: str, total_size: int, block_size: int):  #
    return "{md5:.5}_{total_size}_{block_size}_{{index:0{block_ipad}d}}_{{md5:.5}}".format(
        block_ipad=len(str(total_size // block_size + (0 if (total_size % block_size != 0) else -1))),
        md5=md5,
        total_size=total_size,
        block_size=block_size,
    )


def mask2(md5: str, total_size: int, block_size: int):  #
    s = block_size
    for x in "BKMGTPEZY":
        v, r = divmod(s, 1024)
        if r != 0:
            break
        s = v
    if md5:
        return "{md5:.5}_{block_size}{{index:0{block_ipad}d}}_{{md5:.5}}_{total_size}".format(
            block_ipad=len(str(total_size // block_size + (0 if (total_size % block_size != 0) else -1))),
            md5=md5,
            total_size=total_size,
            block_size=str(s) + x,
        )
    else:
        return "{total_size}{block_size}{{index:0{block_ipad}d}}".format(
            block_ipad=len(str(total_size // block_size + (0 if (total_size % block_size != 0) else -1))),
            total_size=total_size,
            block_size=str(s) + x,
        )


def mask3(md5: str, total_size: int, block_size: int):  #
    s = block_size
    for x in "BKMGTPEZY":
        v, r = divmod(s, 1024)
        if r != 0:
            break
        s = v
    if md5:
        return "{md5:.5}_{block_size}{{index:0{block_ipad}d}}_{total_size}".format(
            block_ipad=len(str(total_size // block_size + (0 if (total_size % block_size != 0) else -1))),
            md5=md5,
            total_size=total_size,
            block_size=str(s) + x,
        )
    else:
        return "{total_size}{block_size}{{index:0{block_ipad}d}}".format(
            block_ipad=len(str(total_size // block_size + (0 if (total_size % block_size != 0) else -1))),
            total_size=total_size,
            block_size=str(s) + x,
        )


def mask4(md5: str, total_size: int, block_size: int):  #
    s = block_size
    for x in "BKMGTPEZY":
        v, r = divmod(s, 1024)
        if r != 0:
            break
        s = v
    block_ipad = len(str(total_size // block_size + (0 if (total_size % block_size != 0) else -1)))
    if md5:
        return "{md5:.5}_{block_size}{{index:0{block_ipad}d}}".format(block_ipad=block_ipad, md5=md5, block_size=str(s) + x)
    else:
        return "{total_size}{block_size}{{index:0{block_ipad}d}}".format(
            block_ipad=block_ipad, total_size=total_size, block_size=str(s) + x
        )


name_fmt = mask1


def set_mask(x):
    global name_fmt
    name_fmt = x


class ChunkWriter(AutoGet):
    chunk_size: int


class ChunkWriterDir(ChunkWriter):
    def __init__(self, chunk_size: int, dir: str):
        self.files = {}
        self.chunk_size = chunk_size
        self.target_dir = dir

    def update(self, o, what, *args):
        if isinstance(o, ChunkLister):
            if self.chunk_size != o.chunk_size:
                pass
            elif what == "in":  #  New chunk started
                # Create new temp file for incoming chunk
                assert o.current_index not in self.files
                from tempfile import NamedTemporaryFile as TempFile

                if self.target_dir:
                    from os.path import isdir

                    if not isdir(self.target_dir):
                        from os import makedirs

                        makedirs(self.target_dir)
                self.files[o.current_index] = TempFile(dir=self.target_dir, delete=None)
            elif what == "out":  # Chunk completed
                #  Closes temp file, stores path
                assert isinstance(o, ChunkLister)
                self.files[o.current_index].close()
                self.files[o.current_index] = self.files[o.current_index].name
            elif what == "data":  # Chunk data received
                assert isinstance(o, ChunkLister)
                # Writes chunk data to temp file
                self.files[o.current_index].write(args[0])
        elif isinstance(o, BlobDescriptor):
            if what == "descriptor_saved":
                # Renames all chunks using descriptor metadata
                desc: "dict[str:object]" = args[0]
                block_size = self.chunk_size
                mask = name_fmt(desc["md5"], desc["size"], block_size)
                for i, c in enumerate(desc["chunks"][block_size]):
                    path = self.files[i]
                    name = mask.format(index=i, md5=c["md5"])
                    self.final_name(path, name)

    def final_name(self, path, name):
        from os.path import split, join
        from os import rename
        from logging import info

        pwd, _ = split(path)
        path2 = join(pwd, name)
        info("Renaming %r to %r", path, path2)
        rename(path, path2)


class ChunkWriterCmd(ChunkWriter):
    """A chunk writer that executes shell commands for each processed chunk.

    This writer handles chunk processing by:
    1. Buffering chunk data in temporary storage (memory/file)
    2. Executing user-provided shell commands when chunks are complete
    3. Supporting range-based filtering for partial processing
    4. Handling both memory and file-based temporary storage

    """

    tmp: IO

    def __init__(self, chunk_size: int, cmd, source_tmp: "None | str" = None, ranges: "None | list[tuple[int, int]]" = None):
        self.cmd = cmd
        self.chunk_size = chunk_size
        self.ranges = ranges
        if source_tmp is not None:
            self.source_tmp = source_tmp

    def _get_source_tmp(self):
        return "file"

    def _get_tmp(self):
        _ = self.source_tmp
        if not _:
            from tempfile import TemporaryFile as TempFile

            return TempFile()
        elif _ == "mem":
            from io import BytesIO

            return BytesIO()
        else:
            return open(_, "wb")

    def allow(self, d: "dict[str, object]"):
        """Check if chunk should be processed based on index ranges."""
        ranges = self.ranges
        if ranges:
            i: int = d["index"]
            for x in ranges:
                if i >= x[0] and i <= x[1]:
                    return True
            return False
        return True

    def update(self, o, what, *args):
        """Handle chunk processing events."""
        if isinstance(o, BlobDescriptor):
            if what == "descriptor_saved":
                self.again(o, args[0])
                pass
        elif isinstance(o, ChunkLister):
            if getattr(self, "chunk_lister", None) != o:
                pass
            elif self.chunk_size != o.chunk_size:
                assert 0
            elif what == "in":
                self.tmp.seek(0)
                self.tmp.truncate()
            elif what == "out":
                assert isinstance(o, ChunkLister)
                d = args[0]
                assert d["size"] == self.tmp.tell()
                self.tmp.seek(0)
                r = self.descriptor["chunks"][self.chunk_size][d["index"]]
                for n in ("md5", "size", "index", "prefix"):
                    if not (d[n] == r[n]):
                        raise RuntimeError("{} not equal {!r} {!r}".format(n, d, r))
                name = self.mask.format(index=d["index"], md5=d["md5"])
                if self.allow(d):
                    self.upload(self.tmp, name)
            elif what == "data":
                assert isinstance(o, ChunkLister)
                self.tmp.write(args[0])

    def again(self, bd: BlobDescriptor, desc: "dict[str, object]"):
        """Re-process descriptor to trigger command execution."""
        total_size = desc["size"]
        block_size = self.chunk_size
        block_count = total_size // block_size + ((total_size % block_size != 0) and 1 or 0)
        _block_ilast = block_count - 1
        self.chunk_lister = cl = ChunkLister(self.chunk_size)
        self.descriptor = desc
        self.current = {}
        self.mask = name_fmt(desc["md5"], desc["size"], block_size)
        cl.observers.append(self)
        buf_size = 64 * 1024
        total_size = 0
        for f in bd.iter_files():
            file_size = 0
            _offset = total_size
            for b in f.iter_chunks(buf_size):
                l = len(b)
                file_size += l
                total_size += l
                cl.update(b)
        cl.get_chunks()  # push the last

    def upload(self, src: IO, name: str):
        """Execute the configured command for a completed chunk."""
        from logging import info
        from shlex import quote
        from subprocess import Popen, check_call, PIPE

        kwa = dict(name=quote(name), size=str(self.chunk_size))
        ckw = dict(shell=True)
        if self.source_tmp == "mem":
            ckw["stdin"] = PIPE
            kwa["file"] = "-"
        elif self.source_tmp:
            kwa["file"] = self.source_tmp
        else:
            ckw["stdin"] = src
            kwa["file"] = "-"
        cmd = self.cmd.format(**kwa)
        info("Calling {!r}".format(cmd))
        environ["FILE"] = kwa["file"]
        environ["NAME"] = name
        environ["SIZE"] = kwa["size"]

        if self.source_tmp == "mem":
            Popen(cmd, **ckw).communicate(src.getvalue())
        else:
            check_call(cmd, **ckw)



import re
import unittest
import os
import tempfile
import subprocess
from pathlib import Path
import shutil


class TestBlobDescriptorCLI(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Create test directory structure
        cls.test_dir = Path(tempfile.mkdtemp())
        cls.other_dir = Path(tempfile.mkdtemp())
        # Create multi-level directory structure
        cls.input_dir = cls.test_dir / "input"
        cls.input_dir.mkdir()
        (cls.input_dir / "subdir").mkdir()
        (cls.input_dir / "subdir2").mkdir()

        # Create test files at different levels
        cls.file1 = cls.input_dir / "file1.bin"
        cls.file2 = cls.input_dir / "subdir" / "file2.bin"
        cls.file3 = cls.input_dir / "subdir2" / "nested" / "file3.bin"
        cls.file3.parent.mkdir()  # Create nested dir

        # Create sample binary files (1MB each)
        for f in [cls.file1, cls.file2, cls.file3]:
            with f.open("wb") as fh:
                fh.write(os.urandom(1024 * 1024))

    @classmethod
    def tearDownClass(cls):
        # Clean up test directory
        shutil.rmtree(cls.test_dir)
        shutil.rmtree(cls.other_dir)

    def _run_command(self, cmd_args):
        """Helper to run commands and print output"""
        cmd = [str(x) for x in (["python", "-m", "blob_descriptor"] + cmd_args)]
        print(f"\n\nExecuting: {' '.join(cmd)}")
        result = subprocess.run(cmd, capture_output=True, text=True, cwd=str(self.test_dir))
        print("Output:")
        print(result.stdout)
        if result.stderr:
            print("Error:")
            print(result.stderr)
        return result

    # @unittest.skip("Enable when testing create command with multi-level dirs")
    def test_create_with_directory_tree(self):
        # Test creating descriptor from directory tree
        result = self._run_command(
            ["create", "--mask", "1", "--cw", "1M,./chunks_1m", "-o", "tree_test.bd", str(self.input_dir)]
        )
        self.assertEqual(result.returncode, 0)
        self.assertTrue((self.test_dir / "tree_test.bd").exists())
        self.assertTrue((self.test_dir / "chunks_1m").exists())
        a = [*(self.test_dir / "chunks_1m").iterdir()]
        # subprocess.run(["find", self.test_dir])
        print(a)
        for x in a:
            self.assertRegex(x.name, r"^\w+_\d+_\d+_\d+_\w+$")
        self.assertTrue(a)

        # Verify all files were included
        self.assertIn("file1.bin", result.stderr)
        self.assertIn("subdir/file2.bin", result.stderr)
        self.assertIn("subdir2/nested/file3.bin", result.stderr)

    # @unittest.skip("Enable when testing create command")
    def test_create_command(self):
        # Test creating a descriptor with chunks
        result = self._run_command(
            [
                "create",
                "--cw",
                "1M,./chunks",
                "--mask",
                "2",
                "--o:dir",
                self.other_dir,
                "--on-saved",
                "python -c \"open('saved.txt', 'w').write(r'{name} {file} {dir}')\"",
                str(self.file1),
                str(self.file2),
            ]
        )
        subprocess.run(["find", self.test_dir])
        subprocess.run(["cat", self.test_dir / "saved.txt"])
        self.assertEqual(result.returncode, 0)
        with (self.test_dir / "saved.txt").open() as f:
            n, f, d = f.read().split()
            self.assertEqual(self.other_dir, Path(d))
            self.assertEqual(Path(f).parent, Path(d))
            self.assertEqual(Path(f).name, n)
        # self.assertTrue((self.test_dir / "test.bd").exists())
        self.assertTrue((self.test_dir / "chunks").exists())

    # @unittest.skip("Enable when testing verify command")
    def test_verify_command(self):
        # First create a descriptor
        self._run_command(["create", "--cw", "1M,1MChunks", "-o", "verify_test.bd", str(self.file1)])
        subprocess.run(["find", self.test_dir])
        # Then verify it
        result = self._run_command(["verify", "verify_test.bd", self.test_dir / "1MChunks", "-d", self.file1.parent])
        self.assertEqual(result.returncode, 0)
        self.assertIn("COMPLETE", result.stdout)

    @unittest.skip("Enable when testing check command")
    def test_check_command(self):
        # First create a descriptor with chunks
        self._run_command(["create", "--cs", "512k", "--cw", "512k,./check_chunks", "-o", "check_test.bd", str(self.file1)])

        # Then check it
        result = self._run_command(["check", "check_test.bd", "./check_chunks"])
        self.assertEqual(result.returncode, 0)
        self.assertIn("OK", result.stdout)

    @unittest.skip("Enable when testing assemble command")
    def test_assemble_command(self):
        # First create a descriptor with chunks
        self._run_command(
            ["create", "--cs", "512k", "--cw", "512k,./assemble_chunks", "-o", "assemble_test.bd", str(self.file1)]
        )

        # Then assemble it
        output_file = self.test_dir / "assembled.bin"
        result = self._run_command(["assemble", "assemble_test.bd", "--sink", str(output_file)])
        self.assertEqual(result.returncode, 0)
        self.assertTrue(output_file.exists())
        self.assertEqual(output_file.stat().st_size, self.file1.stat().st_size)

    @unittest.skip("Enable when testing all commands together")
    def test_full_workflow(self):
        # 1. Create descriptor and chunks
        create_result = self._run_command(
            ["create", "--cs", "512k", "--cw", "512k,./workflow_chunks", "-o", "workflow.bd", str(self.file1), str(self.file2)]
        )
        self.assertEqual(create_result.returncode, 0)

        # 2. Verify
        verify_result = self._run_command(["verify", "workflow.bd", "./workflow_chunks"])
        self.assertEqual(verify_result.returncode, 0)

        # 3. Check
        check_result = self._run_command(["check", "workflow.bd", "./workflow_chunks"])
        self.assertEqual(check_result.returncode, 0)

        # 4. Assemble to new file
        output_file = self.test_dir / "workflow_output.bin"
        assemble_result = self._run_command(["assemble", "workflow.bd", "--sink", str(output_file), "--delete"])
        self.assertEqual(assemble_result.returncode, 0)
        self.assertTrue(output_file.exists())
        self.assertEqual(output_file.stat().st_size, self.file1.stat().st_size + self.file2.stat().st_size)


if __name__ == "__main__":
    unittest.main()



from typing import Any


def list_ranges(indexes):
    # indexes must be sorted, mutable, tempoary list
    r = []
    while indexes:
        x = indexes.pop(0)
        if r:
            a, b = r[-1]
            assert len(r[-1]) == 2
            assert r[-1][0] <= r[-1][1], r[-1]
            assert r[-1][0] >= 0
            # print(x, r[-1])
            if x <= b and x >= a:
                pass
                # r[-1][0] = x
            elif b + 1 == x:
                r[-1][1] = x
            else:
                r.append([x, x])
        else:
            r.append([x, x])
    return r


# Idea by Ben Voigt in:
# https://stackoverflow.com/questions/32869247/a-container-for-integer-intervals-such-as-rangeset-for-c


def sort_condense(ivs):
    if len(ivs) == 0:
        return []
    if len(ivs) == 1:
        if ivs[0][0] > ivs[0][1]:
            return [(ivs[0][1], ivs[0][0])]
        else:
            return ivs
    # eps = sorted(x for xx in (
    #       ((min(iv), False), (max(iv), True)) for iv in ivs)
    #           for x in xx)
    eps = []
    for iv in ivs:
        eps.append((min(iv), False))
        eps.append((max(iv), True))
    eps.sort()
    ret = []
    i = level = 0
    while i < len(eps) - 1:
        if not eps[i][1]:
            level = level + 1
            if level == 1:
                left = eps[i][0]
        else:
            if level == 1:
                if not eps[i + 1][1] and eps[i + 1][0] == eps[i][0] + 1:
                    i = i + 2
                    continue
                right = eps[i][0]
                ret.append((left, right))
            level = level - 1
        i = i + 1
    ret.append((left, eps[len(eps) - 1][0]))
    return ret


def filesizep(s: str):
    if s[0].isnumeric() or s[0].startswith("."):
        q = s.lower()
        if q.endswith("b"):
            q = q[0:-1]
        for i, v in enumerate("kmgtpezy"):
            if q[-1].endswith(v):
                return int(float(q[0:-1]) * (2 ** (10 * (i + 1))))
        return int(q)
    return int(s)


def filesizef2(s):
    x = ""
    t = ["B", "K", "M", "G", "T", "P", "E", "Z", "Y"]
    i = len(t)
    while i > 0:
        i -= 1
        p = i * 10
        b = 1 << p
        if s >= b:
            # x = (x and (x + '.') or '') + str(s>>p) + t[i]
            if x:
                x += str(s >> p) + t[i].lower()
            else:
                x = str(s >> p) + t[i].upper()
            # x = x + str(s>>p) + t[i]
            s %= b
    return x or "0"


class AutoGet:

    def __getattr__(self, name: str) -> object:
        if not name.startswith("_get_"):
            f = getattr(self, f"_get_{name}", None)
            if f:
                setattr(self, name, None)
                v = f()
                setattr(self, name, v)
                return v
        try:
            m = super().__getattr__
        except AttributeError:
            raise AttributeError(f"{self.__class__.__name__} has no attribute {name}") from None
        else:
            return m(name)


def input_file(msg=""):
    from sys import stdin, stderr
    from os.path import exists

    while 1:
        print(f"{msg}: ", file=stderr, flush=True, end="")
        x = stdin.readline().strip()
        if not exists(x) and (x.startswith('"') and x.endswith('"')) or (x.startswith("'") and x.endswith("'")):
            x = x[1:-1]
        if exists(x):
            return x
        print(f"{x!r} not found", file=stderr, flush=True, end="")



import unittest
from blob_descriptor.utils import sort_condense


class TestSortCondense(unittest.TestCase):
    def test_empty_input(self):
        self.assertEqual(sort_condense([]), [])

    def test_single_interval(self):
        self.assertEqual(sort_condense([(1, 3)]), [(1, 3)])

    def test_single_reversed_interval(self):
        self.assertEqual(sort_condense([(3, 1)]), [(1, 3)])

    def test_non_overlapping_intervals(self):
        self.assertEqual(sort_condense([(1, 3), (4, 6), (7, 9)]), [(1, 9)])

    def test_adjacent_intervals(self):
        self.assertEqual(sort_condense([(1, 3), (3, 5), (5, 7)]), [(1, 7)])

    def test_overlapping_intervals(self):
        self.assertEqual(sort_condense([(1, 4), (2, 5), (6, 8)]), [(1, 8)])

    def test_nested_intervals(self):
        self.assertEqual(sort_condense([(1, 10), (2, 5), (3, 4)]), [(1, 10)])

    def test_mixed_order_intervals(self):
        self.assertEqual(sort_condense([(5, 8), (1, 3), (2, 4), (6, 9)]), [(1, 9)])

    def test_reversed_and_normal_intervals(self):
        self.assertEqual(sort_condense([(7, 2), (3, 5), (1, 4)]), [(1, 7)])

    def test_single_point_intervals(self):
        self.assertEqual(sort_condense([(1, 1), (2, 2), (3, 3)]), [(1, 3)])

    def test_consecutive_single_points(self):
        self.assertEqual(
            sort_condense([(1, 1), (2, 2), (3, 3), (4, 4)]),
            [(1, 4)],
        )

    def test_large_numbers(self):
        self.assertEqual(sort_condense([(1000, 2000), (1500, 2500)]), [(1000, 2500)])

    def test_negative_numbers(self):
        self.assertEqual(sort_condense([(-5, -1), (-3, 0), (1, 2)]), [(-5, 2)])

    def test_duplicate_intervals(self):
        self.assertEqual(sort_condense([(1, 3), (1, 3), (1, 3)]), [(1, 3)])

    def test_very_close_intervals(self):
        self.assertEqual(sort_condense([(1, 2), (2, 3), (3, 4), (5, 6)]), [(1, 6)])

    def test_no_intervals(self):
        self.assertEqual(sort_condense([(1, 2), (4, 5), (7, 8)]), [(1, 2), (4, 5), (7, 8)])
        self.assertEqual(sort_condense([(1, 2), (3, 5), (7, 8)]), [(1, 5), (7, 8)])
        self.assertEqual(sort_condense([(1, 2), (2, 5), (7, 8)]), [(1, 5), (7, 8)])


if __name__ == "__main__":
    unittest.main()



from argparse import ArgumentParser
from .check import Check
from .verify import Verify
from .assemble import Assemble
from .create import Create
from .main import Main


class App(Main):
    """Main application class."""

    def init_argparse(self, argp: ArgumentParser) -> None:
        argp.prog = "blob_descriptor"
        return super().init_argparse(argp)

    def sub_args(self) -> object:
        """Register all subcommands."""
        yield Verify(), {"name": "verify", "help": "Check files"}
        yield Check(), {"name": "check", "help": "Check chunks"}
        yield Create(), {"name": "create", "help": "Create descriptor"}
        yield Assemble(), {"name": "assemble", "help": "Assemble files"}


def main():
    """CLI entry point."""
    App().main()


if __name__ == "__main__":
    main()



from os.path import relpath, basename
from urllib.parse import urlparse
from . import BlobDescriptor, ChunkWriterDir, ChunkWriterCmd, FileSource, URLSource
from .main import Main, arg, flag
from .utils import filesizep


def abs2rel(path, base):
    try:
        return relpath(path, base)
    except ValueError:
        return basename(path)


class Observer:
    def __init__(self):
        self.when = {}

    def update(self, o, what, *args):
        f = self.when.get(what)
        f and f(*args)


class Create(Main):
    files: "list[str]" = arg("files to add", nargs="+")
    duplicate: bool = flag("link duplicate files", default=None)
    dry_run: bool = flag("dry-run", "not a trial run", default=False)
    base_dir: str = flag("b", "sub directory of files added in descriptor", default="")
    out: str = flag("o", "out", "save descriptor to FILE", metavar="FILE")
    mask: int = flag("m", "mask", "chunks name mask", choices=[1, 2, 3, 4])
    cmd_on_saved: str = flag("on-saved", "run CMD after descriptor file is created", metavar="CMD")
    chunk_sizes: str = flag(
        "chunk-size",
        "cs",
        "include in descriptor block SIZE e.g. '32k', '2M', '1g'",
        metavar="SIZE",
    )
    chunk_write: "list[str]" = flag(
        "chunk-write", "cw", "write block SIZE, in DIR e.g. '25m,/tmp/25MB'", metavar="SIZE,DIR", default=[]
    )
    chunk_call: "list[str]" = flag(
        "chunk-call", "cc", "run CMD in every block SIZE created at TMP file", metavar="SIZE,CMD,TMP", default=[]
    )
    o_stemf: str = flag("o:stemf", "descriptor filename stem part mask", metavar="MASK", default="")
    o_dir: str = flag("o:dir", "descriptor file destination DIRectory", metavar="DIR", default="")

    def start(self) -> None:
        from os import environ
        from logging import basicConfig
        from sys import stdin

        format = environ.get("LOG_FORMAT", "%(levelname)s: %(message)s")
        level = environ.get("LOG_LEVEL", "INFO")
        basicConfig(format=format, level=level)

        bd = BlobDescriptor()
        if self.duplicate is not None:
            bd.no_duplicates = not self.duplicate

        if self.mask:
            from . import set_mask, mask1, mask2, mask3, mask4

            set_mask((mask1, mask2, mask3, mask4)[int(self.mask) - 1])

        if x := self.chunk_sizes:
            bd.chunk_writers.extend([filesizep(x) for x in x.split(",") if x])

        for x in self.chunk_write:
            size, wdir = [v.strip() for v in x.split(",")]
            bd.chunk_writers.append(ChunkWriterDir(filesizep(size), wdir))

        for x in self.chunk_call:
            size, tmp, cmd, *etc = [v.strip() for v in x.split(",")]
            # print("CC", x, (size, tmp, cmd, *etc))
            cwc = ChunkWriterCmd(filesizep(size), cmd, tmp)
            # print(x)
            if etc:
                cwc.ranges = ranges = []
                for r in etc:
                    s, _, e = r.partition("-")
                    if e:
                        ranges.append((int(s), int(e)))
                    else:
                        ranges.append((int(s), int(s)))
                # print(x, cwc.ranges)
            else:
                pass
            bd.chunk_writers.append(cwc)
        obs = Observer()
        if self.cmd_on_saved:
            from shlex import quote
            from subprocess import check_call
            from os.path import split, abspath

            def fn(desc, path, *args):
                pwd, name = split(abspath(path))
                ckw = dict(shell=True)
                environ["NAME"] = name
                environ["FILE"] = path
                environ["DIR"] = pwd
                cmd = self.cmd_on_saved.format(**dict(name=quote(name), file=quote(path), dir=quote(pwd)))
                check_call(cmd, **ckw)

            obs.when["descriptor_saved"] = fn
        if len(obs.when) > 0:
            bd.observers.append(obs)
        # ADD files
        from os.path import isdir

        for f in self.files:
            if not f:
                pass
            elif f.startswith("stdin:///"):
                bd.files.append(FileSource(stdin.buffer, path=urlparse(f)[2].strip("/")))
            elif "://" in f:
                bd.files.append(URLSource(f))
            elif "-" == f:
                for s in stdin:
                    f = s.strip()
                    if not f:
                        continue
                    if self.base_dir:
                        bd.add_file(f, abs2rel(f, self.base_dir))
                    else:
                        bd.add_file(f)
            elif isdir(f):
                bd.add_tree(f)
            elif self.base_dir:
                bd.add_file(f, abs2rel(f, self.base_dir))
            elif f:
                bd.add_file(f)

        p = {}
        if self.o_dir:
            p["dir"] = self.o_dir
        if self.o_stemf:
            p["stemf"] = self.o_stemf
        if self.out:
            p[""] = self.out
        if len(bd.files) > 1:
            bd.files.sort(key=lambda x: x.path)

        if self.dry_run is False:
            bd.save(p)


(__name__ == "__main__") and Create().main()



from shutil import rmtree
import unittest
import os
import tempfile
import subprocess
from pathlib import Path


class TestBlobDescriptorCLI(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Create test files
        cls.test_dir = tempfile.mkdtemp()
        cls.file1 = os.path.join(cls.test_dir, "apple.bin")
        cls.file2 = os.path.join(cls.test_dir, "banana.bin")

        # Create sample binary files (1MB each)
        with open(cls.file1, "wb") as f:
            f.write(os.urandom(1024 * 1024))
        with open(cls.file2, "wb") as f:
            f.write(os.urandom(1024 * 1024))

    @classmethod
    def tearDownClass(cls):
        # Clean up test directory
        rmtree(cls.test_dir)

    def _run_command(self, cmd_args):
        """Helper to run commands and print output"""
        cmd = ["python", "-m", "blob_descriptor"] + cmd_args
        print(f"\n\nExecuting: {' '.join(cmd)}")
        result = subprocess.run(cmd, capture_output=True, text=True, cwd=self.test_dir)
        print("Output:")
        print(result.stdout)
        if result.stderr:
            print("Error:")
            print(result.stderr)
        return result

    # @unittest.skip("Enable when testing create command")
    def test_create_command(self):
        # Test creating a descriptor with chunks
        result = self._run_command(["create", "--cs", "1m", "--cw", "512k,./chunks", "-o", "test.bd", self.file1, self.file2])
        self.assertEqual(result.returncode, 0)
        subprocess.run(["find", self.test_dir])
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "test.bd")))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, "chunks")))

    # @unittest.skip("Enable when testing verify command")
    def test_verify_command(self):
        # First create a descriptor
        self._run_command(["create", "--cs", "512k", "-o", "verify_test.bd", self.file1])

        # Then verify it
        result = self._run_command(["verify", "verify_test.bd"])
        self.assertEqual(result.returncode, 0)
        self.assertIn("COMPLETE", result.stdout)

    @unittest.skip("Enable when testing check command")
    def test_check_command(self):
        # First create a descriptor with chunks
        self._run_command(["create", "--cs", "512k", "--cw", "512k,./check_chunks", "-o", "check_test.bd", self.file1])
        subprocess.run(["find", self.test_dir])
        # Then check it
        result = self._run_command(["check", "check_test.bd", "./check_chunks"])
        self.assertEqual(result.returncode, 0)
        self.assertIn("OK", result.stdout)

    # @unittest.skip("Enable when testing assemble command")
    def test_assemble_command(self):
        # First create a descriptor with chunks
        self._run_command(["create", "--cs", "512k", "--cw", "512k,./assemble_chunks", "-o", "assemble_test.bd", self.file1])

        # Then assemble it
        output_file = os.path.join(self.test_dir, "assembled.bin")
        result = self._run_command(["assemble", "assemble_test.bd", "--sink", output_file, "-s", "assemble_chunks"])
        self.assertEqual(result.returncode, 0)
        self.assertTrue(os.path.exists(output_file))
        self.assertEqual(os.path.getsize(output_file), os.path.getsize(self.file1))

    # @unittest.skip("Enable when testing all commands together")
    def test_full_workflow(self):
        # 1. Create descriptor and chunks
        create_result = self._run_command(
            ["create", "--cs", "512k", "--cw", "512k,./workflow_chunks", "-o", "workflow.bd", self.file1, self.file2]
        )
        self.assertEqual(create_result.returncode, 0)

        # 2. Verify
        verify_result = self._run_command(["verify", "workflow.bd", "./workflow_chunks"])
        self.assertEqual(verify_result.returncode, 0)

        # 3. Check
        check_result = self._run_command(["check", "workflow.bd", "./workflow_chunks"])
        self.assertEqual(check_result.returncode, 0)

        # 4. Assemble to new file
        output_file = os.path.join(self.test_dir, "workflow_output.bin")
        assemble_result = self._run_command(
            ["assemble", "workflow.bd", "--sink", output_file, "--delete", "-s", "./workflow_chunks"]
        )
        self.assertEqual(assemble_result.returncode, 0)
        self.assertTrue(os.path.exists(output_file))
        self.assertEqual(os.path.getsize(output_file), os.path.getsize(self.file1) + os.path.getsize(self.file2))


if __name__ == "__main__":
    unittest.main()



def read_descriptor(file):
    from re import match, compile as regex
    from ast import literal_eval

    if hasattr(file, "read"):
        desc = h.read()
    elif match(r"https?\:\/\/.+", file):
        from requests import get

        with get(file) as r:
            desc = r.content
    else:
        with open(file, "rb") as h:
            desc = h.read()

    if desc.startswith(b"files"):
        lines = desc.decode("UTF-8").splitlines()
        bd = {"files": [], "chunks": {}}
        _files = _chunks = _chunk_size = None
        m_chunks = regex(r"^\s*(\w+)(\@\d+)?(\:\d+)?\s+(.+)\s*$")
        m_files = regex(r"^\s*(\w+)(@\d+)?(:\d+)\s+(.+)\s*$")
        for x in lines:
            x = x.strip()
            if x == "files":
                _files = _chunks = None
                _files = bd["files"]
                # print(x, "files", _files)
            elif x.startswith("chunks:"):
                _files = _chunks = None
                _, size = x.split(":")
                _chunk_size = int(size)
                _chunks = bd["chunks"].setdefault(_chunk_size, [])
                # print(x, "chunks", _chunks)
            elif x:
                # print(x, bool(_files), bool(_chunks))
                assert _files is not None or _chunks is not None
                if _files is not None:
                    m = m_files.match(x)
                    checksum = m.group(1)
                    offset = m.group(2)
                    size = m.group(3)
                    path = m.group(4)
                    size = None if size is None else int(size.lstrip(":"))
                    if path == ".":
                        bd["md5"] = checksum
                        bd["size"] = size
                    else:
                        _files.append(
                            dict(
                                path=path,
                                offset=(
                                    (0 if len(_files) < 1 else _files[-1]["offset"] + _files[-1]["size"])
                                    if offset is None
                                    else int(offset.lstrip("@"))
                                ),
                                size=size,
                                md5=checksum,
                            )
                        )
                elif _chunks is not None:
                    m = m_chunks.match(x)
                    checksum = m.group(1)
                    index = m.group(2)
                    size = m.group(3)
                    prefix = m.group(4)
                    entry = dict(
                        prefix=prefix,
                        md5=checksum,
                    )

                    if index is None:
                        entry["index"] = len(_chunks)
                        _chunks.append(entry)
                    else:
                        index = int(index.lstrip("@"))
                        entry["index"] = index
                        diff = (index + 1) - len(_chunks)
                        if diff > 0:
                            _chunks.extend([None] * diff)
                        _chunks[index] = entry
                    if size is None:
                        # s = (entry["index"] + 1) * _chunk_size
                        remain = bd["size"] % _chunk_size
                        last_index = (bd["size"] // _chunk_size) + (1 if remain != 0 else 0) - 1

                        if entry["index"] == last_index:
                            size = remain if remain != 0 else _chunk_size
                        else:
                            size = _chunk_size
                    else:
                        size = int(size.lstrip(":"))
                    entry["size"] = size

        return bd
    else:
        return literal_eval(desc.decode("UTF-8"))


def _iter_descriptor(bd, file_offset=None, chunk_size=None, chunk_index=None):
    yield f"files"
    yield f'{bd["md5"]}:{bd["size"]} .'
    for x in bd["files"]:
        yield f'{x["md5"]}' + (
            "" if file_offset is False or x["offset"] is None else "@%d" % x["offset"]
        ) + f':{x["size"]} {x["path"]}'
    for size, chunks in bd["chunks"].items():
        yield f"chunks:{size}"
        for x in chunks:
            yield f'{x["md5"]}' + ("" if chunk_index is False or x["index"] is None else "@%d" % x["index"]) + (
                "" if chunk_size is False or x["size"] is None else ":%d" % x["size"]
            ) + f' {x["prefix"]}'


def write_descriptor(bd: dict, file: str, compact=None, **kwargs):
    if compact is None:
        with open(file, "w") as h:
            for x in _iter_descriptor(bd, **kwargs):
                h.write(f"{x}\n")
        bd2 = read_descriptor(file)
        assert bd == bd2
        # try:
        #     bd2 = read_descriptor(file)
        #     assert bd == bd2
        # except AssertionError:
        #     from unittest import TestCase

        #     import pprint

        #     pprint.pprint(bd)
        #     pprint.pprint(bd2)
        #     TestCase().assertDictEqual(bd, bd2)

    else:
        if compact:
            b = repr(bd).encode("UTF-8")
        else:
            from pprint import pformat

            b = pformat(bd).encode("UTF-8")
        with open(file, "wb") as h:
            h.write(b)
    return True


# ['files', 'md5', 'size', 'chunks']

# bd['files'][0].keys()
# dict_keys(['md5', 'size', 'path', 'offset'])

# bd['chunks'].keys()
# dict_keys([524288, 26214400])

# bd['chunks'][524288][0].keys()
# dict_keys(['md5', 'prefix', 'size', 'index'])

# bd['chunks'][26214400][0].keys()
# dict_keys(['md5', 'prefix', 'size', 'index'])

# [ (x,type(bd[x])) for x in bd.keys() ]
# [('files', list), ('md5', str), ('size', int), ('chunks', dict)]



from .verify import Verify
from .construct import SerialConstructor
from .finder import ChunkFinder
from .utils import sort_condense


class Check(Verify):

    def CHECK(self, path: str, finder: "ChunkFinder") -> None:
        from hashlib import md5 as mdnew

        const = SerialConstructor(path)
        const.pwd = self.output_dir
        const.chunk_finders.append(finder)
        print("chunks", const.descriptor["chunks"].keys())

        total_size: int = int(const.descriptor["size"])
        block_sizes: set[int] = set(x["block_size"] for x in const.sources)
        has_ranges = sort_condense(list(const.available_ranges()))
        print("available_ranges", has_ranges, *const.descriptor.keys())
        non_ranges = sort_condense(list(const.non_available_ranges()))
        print("non_available_ranges", non_ranges)

        bmap = dict((s, {}) for s in block_sizes)

        for s in block_sizes:
            total_blocks = total_size // s + (1 if (total_size % s) else 0)
            bmap[s]["index_map"] = index_map = dict((i, None) for i in range(total_blocks))
            print(f"block_size: {s}b x{total_blocks}")
            for a, b in non_ranges:
                d = b - a
                assert d > 0
                last_index = (b // s) + (1 if (b % s) else 0)
                first_index = a // s
                print("\t", sorted(set([first_index, last_index])))
                assert first_index <= last_index
                i = first_index
                while i <= last_index:
                    index_map[i] = False
                    i += 1

        r = []
        for f in const.chunk_finders:
            for x in f.all():
                if not const.descriptor["md5"].startswith(x["checksum"]):
                    continue
                # print(
                #     x["path"],
                #     x["part_index"],
                #     x["block_size"] == x["part_size"],
                # )
                s = x["block_size"]
                bmap[s]["index_map"][x["part_index"]] = False
                if x["block_size"] == x["part_size"]:
                    r.append([x["part_index"], x["part_index"] + 1])
                    if s in bmap:
                        h = x.get("part_checksum")
                        if not h:
                            h = const.descriptor["chunks"][s][x["part_index"]]["md5"]
                        # if h:
                        #     md = mdnew()
                        #     md.update(open(x["path"], "rb").read())
                        #     if not md.hexdigest().startswith(h):
                        #         continue
                        #     print(x["path"], "hexdigest", h)

                        # TODO: check hash
                        bmap[s]["index_map"][x["part_index"]] = True
                    pass
        print("\t", sort_condense(sorted(r)))
        for s, m in bmap.items():
            print("bmap:", s, [k for (k, v) in m["index_map"].items() if v is False])

        # const.check()

    def start(self) -> None:
        finder = self.find_chunks()
        for f in self.descriptor_files():
            self.VERIFY(f, finder)

        a = {}
        for x in finder.all():
            p = a.get(x["checksum"])
            if p is None:
                p = a[x["checksum"]] = []
            p.append(x)
        for f in self.descriptor_files():
            self.CHECK(f, finder)



from .verify import Verify
from .utils import input_file, sort_condense
from .main import flag
from .finder import ChunkFinder


class Assemble(Verify):

    bxb: bool = flag("bxb", help="Use block by block method for reading urls", default=False)
    delete: bool = flag("Delete finished chunks")
    max_sectors: int = flag("Max sectors to write", metavar="N")
    rm_extra_blocks: bool = flag("Remove uneeded blocks", default=False)
    sink: str = flag("sink", help="Write blob to SINK", metavar="SINK")
    dry_run: bool = flag("dry-run", "not a trial run", default=False)

    def start(self) -> None:

        from os.path import exists
        from .construct import RequestSectorReader, HtttpRangeSectorReader

        # print(app.config_path)
        def get_params():
            p = {}
            if self.bxb:  # block by block
                p["url_sector_reader"] = HtttpRangeSectorReader
            else:
                p["url_sector_reader"] = RequestSectorReader
            if self.check_hash:
                p["check_hash"] = True
            if self.delete:
                p["delete_finished_blocks"] = True
            if self.max_sectors:
                p["max_sectors"] = self.max_sectors
            if self.output_dir == ":":
                self.output_dir = input_file("Enter output dir")
            return p

        # The Chunks
        params = get_params()
        finder = self.find_chunks()

        # print(self.descriptor)
        # print(self.chunks)
        # print(self.config_path)
        # print(self.chunk_paths.keys())

        if self.sink:
            f = self.sink
            if f in (".", ""):
                sink = None
            elif f == "-":
                from sys import stdout
                from os import fdopen

                # sink = stdout.buffer
                sink = fdopen(stdout.fileno(), "wb", closefd=False)
            else:
                if exists(f):
                    raise RuntimeError(f"Not over writing {f:r}")
                else:
                    sink = open(f, "wb")
            if sink:
                with sink:
                    for f in self.descriptor_files():
                        self.write_to_file(f, finder, params, sink)

        else:
            for f in self.descriptor_files():
                self.write_to_dir(f, finder, params)

    def write_to_file(self, path: str, finder: ChunkFinder, params: dict, sink):
        from .construct import Sink

        const = Sink(path)
        const.chunk_finders.append(finder)
        const.construct(sink, **params)

    def write_to_dir(self, path: str, finder: ChunkFinder, params: dict):
        # OUT: Write to directory
        from .construct import SerialConstructor
        from os.path import exists

        const = SerialConstructor(path)
        const.pwd = self.output_dir
        const.chunk_finders.append(finder)
        const.construct(**params)
        if self.rm_extra_blocks:
            from os import unlink

            total_size = const.descriptor["size"]
            ranges = sort_condense(list(const.available_ranges()))
            for v in const.sources:
                for s, e in ranges:
                    b_start = v["offset"]
                    b_tail_size = total_size % v["block_size"]
                    b_count = total_size // v["block_size"] + ((b_tail_size != 0) and 1 or 0)
                    b_index = v["part_index"]
                    b_end0 = b_start + (b_tail_size if (b_count - 1) == b_index else v["block_size"])
                    b_end = min(total_size, b_start + v["block_size"])
                    try:
                        assert b_end == b_end0
                    except Exception:
                        print(b_end, b_end0)
                        raise
                    path = v.get("path")
                    if b_start >= s and b_end <= e and path and exists(path):
                        print("RM {}-{} {}-{} {!r}".format(s, e, b_start, b_end, path))
                        if self.dry_run is False:
                            unlink(path)


(__name__ == "__main__") and Assemble().main()



class ChunkFinder(object):
    from re import compile as compile_re

    refnchunk1 = compile_re(
        r"""(?ix)
        (?P<checksum>[0-9a-zA-Z]+)[-_]
        (?P<size>\d+)[\._]+
        (?P<block_size>\d+)[-_]
        (?P<part_index>\d+)[-_\.]
        (?P<part_checksum>\w{5,5})
        (?:\.\w{3,4})?"""
    )
    refnchunk3 = compile_re(
        r"""(?ix)
        (?P<checksum>[0-9a-zA-Z]+)_(?P<size>\d+)
        (?:_s(?P<start>[0-9a-zA-Z]+))
        (?:_e(?P<end>[0-9a-zA-Z]+))?
        (?:\.\w{3,4})?"""
    )
    refnchunk2 = compile_re(
        r"(?ix)"
        r"(?P<checksum>[0-9a-zA-Z]+)"
        r"[-_](?P<block_size>\d+[BMGT])(?P<part_index>\d+)"
        r"(?:[-_](?P<part_checksum>\w{5,5}))?"
        r"[-_](?P<size>\d+)"
        r"(?:\.\w{3,4})?"
    )
    # 7149a_15M122_1b476_4663073622
    re_url = compile_re(r"(?P<scheme>https?|ftp)\:\/\/.+")

    def __init__(self):
        self.map_sizes: "dict[int, dict[str, dict[int, dict[int, list[dict]]]]] " = {}
        self.map_parts = {}

    def iter_url_name(self, url):
        from urllib.parse import urlsplit, parse_qs

        v = urlsplit(url)
        if v[3]:
            q = parse_qs(v[3])
            if q.get("filename") and q.get("disposition"):
                # https://s334myt.storage.yandex.net/...&filename=..&disposition=..
                yield q["filename"][0]
        if v[4]:  # fragment
            yield v[4]
        if v[2]:  # path
            v = v[2].rpartition("/")
            if v[2]:
                yield v[2]

    def match(self, name, pwd):
        m = self.re_def(name)
        # print("match", name, m)
        if not m:
            return None
        h = {"basename": name}
        if pwd:
            from os.path import join

            h["path"] = join(pwd, name)
            h["pwd"] = pwd
            from os import stat

            h["part_size"] = stat(h["path"]).st_size
        for k, v in m.groupdict().items():
            if k in ("size", "block_size", "part_index"):
                if v[-1].isalpha():
                    for i, x in enumerate("bkmgtpezy"):
                        if v[-1].lower().endswith(x):
                            h[k] = int(v[0:-1]) * (2 ** (10 * i))
                            break
                else:
                    h[k] = int(v, base=10)
            elif k in ("group", "id"):
                h[k] = v.replace("_", ":").replace(".", ":").replace("-", ":").replace("::", ":")
            else:
                h[k] = v
        return h

    def match_file(self, path, pwd=None):
        if pwd:
            name = path
        else:
            from os.path import split

            pwd, name = split(path)
        if name:
            return self.match(name, pwd)

    def re_def(self, name):
        return self.refnchunk1.search(name) or self.refnchunk2.search(name) or self.refnchunk3.search(name)

    def match_url(self, url, name=None):
        m = None
        if name:
            m = self.re_def(name)
        else:
            for name in self.iter_url_name(url):
                if name:
                    m = self.re_def(name)
                if m:
                    break
        if not m:
            m = self.re_def(url)
        if not m:
            from logging import warn

            warn("Bad URL %r", url)
            return None
        h = {"url": url}
        for k, v in m.groupdict().items():
            if k in ("size", "block_size", "part_index", "start", "end"):
                if v[-1].isalpha():
                    for i, x in enumerate("bkmgtpezy"):
                        if v[-1].lower().endswith(x):
                            h[k] = int(v[0:-1]) * (2 ** (10 * i))
                            break
                else:
                    h[k] = int(v, base=10)
            elif k in ("group", "id"):
                h[k] = v.replace("_", ":").replace(".", ":").replace("-", ":").replace("::", ":")
            else:
                h[k] = v
        # print(h)
        return h

    def accept(self, m):
        return m

    def check_url(self, url):
        m = self.accept(self.match_url(url))
        m and self.add(m)

    def check_file(self, *args):
        m = self.accept(self.match_file(*args))
        m and self.add(m)
        # print('check_file', args, m)

    def search(self, *args):
        from os import listdir

        col = self.map_sizes  # size map_sizes
        for pwd in args:
            m = self.re_url.match(pwd)
            if m:
                self.check_url(pwd)
            else:
                for n in listdir(pwd):
                    self.check_file(n, pwd)
        return self

    def add(self, m):
        try:
            if "block_size" in m:
                (
                    self.map_sizes.setdefault(m["size"], {})
                    .setdefault(m["checksum"], {})
                    .setdefault(m["block_size"], {})
                    .setdefault(m["part_index"], [])
                ).append(m)
            else:
                (self.map_parts.setdefault(m["size"], {}).setdefault(m["checksum"], {}).setdefault(m["start"], [])).append(m)
        except:
            print("Error adding {!r}".format(m))
            raise

    def items(self, md5: str, size: int, part_len: "None | int" = None):
        size_map = self.map_sizes.get(size) or {}
        for md5_prefix in size_map:
            if not md5.startswith(md5_prefix):
                continue
            md5_map = size_map[md5_prefix]
            for block_size in md5_map:
                if part_len and block_size != part_len:
                    continue
                block_size_map = md5_map[block_size]
                for block_index in block_size_map:
                    parts_list = block_size_map[block_index]
                    for part in parts_list:
                        part["offset"] = block_index * block_size
                        yield part
        part_map = self.map_parts.get(size, "")
        for md5_prefix in part_map:
            if not md5.startswith(md5_prefix):
                continue
            md5_map = part_map[md5_prefix]
            for start in md5_map:
                start_map = md5_map[start]
                for part in start_map:
                    part["offset"] = part["start"]
                    yield part

    def all(self):
        for size, size_map in self.map_sizes.items():
            for md5, md5_map in size_map.items():
                for block_size, block_size_map in md5_map.items():
                    for block_index, parts_list in block_size_map.items():
                        for part in parts_list:
                            yield part

    def make_map_file(self, pwd):
        if not pwd:
            from tempfile import gettempdir

            pwd = gettempdir()
        from os.path import join

        for size, size_map in self.map_sizes.items():
            for md5, md5_map in size_map.items():
                maph = {"length": size}
                parts = maph["parts"] = []
                for block_size, block_size_map in md5_map.items():
                    for block_index, parts_list in block_size_map.items():
                        for part in parts_list:
                            s = block_index * block_size
                            parts.append((s, s + block_size, part["path"]))
                name = md5[:5] + "_" + str(size) + ".parts.json"
                ext = ".bin"
                map_file = join(pwd, name + ext)
                with open(map_file, "w") as f:
                    from json import dump

                    dump(maph, f, sort_keys=True, indent=4)

    def make_desc_file(self, pwd):
        if not pwd:
            from tempfile import gettempdir

            pwd = gettempdir()
        from os.path import join

        for size, size_map in self.map_sizes.items():
            for md5, md5_map in size_map.items():
                maph = {"size": size, "md5": md5}
                chunks = maph["chunks"] = {}
                files = maph["files"] = []
                name = md5[:5] + "_" + str(size) + "_desc"
                ext = ".bin" and ""
                files.append(dict(md5=md5, offset=0, size=size, path=md5[:5] + "_" + str(size)))
                for block_size, block_size_map in md5_map.items():
                    for block_index, parts_list in block_size_map.items():
                        parts = chunks.setdefault(block_size, [])
                        for part in parts_list:
                            d = dict(
                                index=part["part_index"],
                                md5=part["part_checksum"],
                                size=part["size"],
                            )
                            for k in ("url", "path"):
                                v = part.get(k)
                                if v:
                                    d[k] = v
                            parts.append(d)
                map_file = join(pwd, name + ext)
                with open(map_file, "wb") as f:
                    from pprint import pformat

                    b = pformat(maph).encode("UTF-8")
                    f.write(b)
